//
//  CLPlacemark+Extensions.swift
//  WherePrayerUtils
//
//  Created by Jeff Shulman on 10/12/17.
//  Copyright © 2017 Lighthouse Woods LLC All rights reserved.
//

import Foundation
import Contacts
import CoreLocation

public extension CLPlacemark {
    /// Returns true if this is a landmark. Assume landmarks have the same name as in areasOfInterest
    func isLandmark() -> Bool {
        if let name = name, let areasOfInterest = areasOfInterest, areasOfInterest.contains(name) {
            return true
        }
        
        return false
    }
    
    /// Returns if this is a "known" placemark. Either it is in the water or a country
    func isKnown() -> Bool {
        return isoCountryCode != nil || inlandWater != nil || ocean != nil
    }
    
    /// Returns the text address for the actual location of the placemark formatted as a postal address
    func fullAddressString() -> String {
        if let wellKnownAddress = wellKnownAddressString() {
            return wellKnownAddress
        }
        
        if let postalAddress = postalAddress {
            return CNPostalAddressFormatter().string(from: postalAddress)
        }
        
        return internalAddressString(formatStr: "In %@ %@ %@")
    }
    
    /// Returns the text address street and city of the placemark
    func shortAddressString() -> String {
        if let wellKnownAddress = wellKnownAddressString() {
            return wellKnownAddress
        }
        
        if let postalAddress = postalAddress {
            if !postalAddress.street.isEmpty && !postalAddress.city.isEmpty && !postalAddress.state.isEmpty && !postalAddress.country.isEmpty {
                return String(format: "*SHORT.ADDRESS.FORMAT*", postalAddress.street, postalAddress.city, postalAddress.state, postalAddress.country)
            }
            else if !postalAddress.street.isEmpty {
                return postalAddress.street
            }
            else if !postalAddress.city.isEmpty {
                return postalAddress.city
            }
            else if !postalAddress.state.isEmpty {
                return postalAddress.state
            }
            else if !postalAddress.country.isEmpty {
                return postalAddress.country
            }
            else {
                return "Unknown"
            }
        }
        else {
            return "Unknown"
        }
    }
    
    /// Returns the added to address of the placemark
    func addedAddressString() -> String {
        let internalAddress = internalAddressString(formatStr: "%@ %@ %@")
        return String(format: "Added to prayer near %@", internalAddress)
    }

    /// Returns the nearby address of the placemark
    func nearAddressString() -> String {
        return internalAddressString(formatStr: "Near %@ %@ %@")
    }
    
    /// Returns the placemark as key archived data
    func archiveData() -> Data {
        let archiver = NSKeyedArchiver(requiringSecureCoding: true)
        encode(with: archiver)
        archiver.finishEncoding()
        
        return archiver.encodedData
    }
    
    private func wellKnownAddressString() -> String? {
        if let ocean = ocean {
            // In an ocean
            return String(format: "In the %@", ocean)
        }
        
        if let inlandWater = inlandWater {
            // Inland water body
            return String(format: "In the %@", inlandWater)
        }
        
        if isLandmark() {
            return name
        }
        
        return nil
    }
    
    // <city> <state> <country>
    private func internalAddressString( formatStr: String ) -> String {
        if let wellKnownAddress = wellKnownAddressString() {
            return wellKnownAddress
        }
        
        let cityLocation = locality ?? administrativeArea ?? country ?? ""
        let stateLocation = ((administrativeArea != nil) && (cityLocation != administrativeArea!)) ? administrativeArea! : (country ?? name ?? "Unknown")
        let countryLocation = country != nil && (cityLocation != country!) && (stateLocation != country) ? country! : ""
        return String(format: formatStr, cityLocation, stateLocation, countryLocation )
    }
}
