//
//  AddressDTO.swift
//  WherePrayerModel
//
//  Created by Jeff Shulman on 12/26/17.
//  Copyright © 2017 Lighthouse Woods LLC All rights reserved.
//
//  Address data transfer object

import Foundation
import CloudKit
import Contacts
import MapKit
import RealmSwift

public struct AddressFlags : OptionSet {
    public let rawValue: Int
    
    public init(rawValue: Int) {
        self.rawValue = rawValue
    }
    
    public static let isDeleted  = AddressFlags(rawValue: 1 << 0)
}

final public class AddressDTO: NSObject {
    public var ckSystemFields: Data?        // CloudKit system fields for this record

    // Address items
    public var addressString: String?       // We use this when searching
    public var flags: AddressFlags = []     // Flags
    public var id: String                   // uuidString
    public var mapImageID: String?          // uuidString
    public var keywords: String?
    public var name: String
    public var placemark: CLPlacemark? {
        didSet {
            addressString = placemark?.fullAddressString() ?? "Unknown"
        }
    }
    public var serverModDate: Date?
    
    public weak var mapImageFetchOperation: Operation?
    
    public override var debugDescription: String {
            return """
            AddressDTO
                id: \(id)
                ckSystemFields: \(String(describing: ckSystemFields))
                flags: \(String(describing: flags))
                mapImageID: \(String(describing: mapImageID)))
                keywords: \(String(describing: keywords))
                name: \(name)
                placemark: \(String(describing: placemark))
                addressString: \(String(describing: addressString))
                serverModDate: \(String(describing: serverModDate))\n
            """
    }
    
    public var calloutMapDistanceInMeters: CLLocationDistance {
        get {
            return Constants.Maps.ImageDistance.PrivatePrayer
        }
    }
    
    public var coordinates: CLLocationCoordinate2D? {
        get {
            return placemark?.location?.coordinate
        }
    }
    
    public var location: CLLocation? {
        get {
            return placemark?.location
        }
    }
    
    public var mapImageWidth: Double {
        get {
            return Double(Constants.Maps.AddressMapWidth)
        }
    }

    public var displayAddress: String {
        get {
            if addressString != nil && !addressString!.isEmpty {
                return addressString!
            }

            return placemark?.fullAddressString() ?? "Unknown"
        }
    }

    public var shortAddressString: String {
        return placemark?.shortAddressString() ?? "Unknown"
    }
    
    // Initializer for creating a new address
    public override init() {
        self.ckSystemFields = nil
        self.id = UUID().uuidString
        self.flags = []
        self.keywords = nil
        self.name = ""
        self.placemark = nil
        self.addressString = nil

        super.init()
    }

    public init(address: AddressObject) {
        self.ckSystemFields = address.ckSystemFields as Data?
        self.id = address.id
        self.flags = AddressFlags(rawValue: address.flags)
        self.keywords = address.keywords
        self.name = address.name
        self.placemark = address.placemark!
        self.addressString = address.addressString!
        self.mapImageID = address.mapImage?.id
        self.serverModDate = address.serverModDate
        
        super.init()
    }
    
    // Copy constructor
    public init( addressDTO: AddressDTO ) {
        self.ckSystemFields = addressDTO.ckSystemFields
        self.flags = addressDTO.flags
        self.id = addressDTO.id
        self.keywords = addressDTO.keywords
        self.name = addressDTO.name
        self.addressString = addressDTO.addressString
        if let placemark = addressDTO.placemark {
            self.placemark = placemark.copy(with: nil) as? CLPlacemark
        }
        self.serverModDate = addressDTO.serverModDate
        
        super.init()
    }
    
    public init(id: String, name: String, placemark: CLPlacemark? = nil, keywords: String? = nil, serverModDate: Date? = nil, flags: AddressFlags = []) {
        self.ckSystemFields = nil
        self.flags = flags
        self.id = id
        self.keywords = keywords
        self.name = name
        self.addressString = placemark?.fullAddressString() ?? "Unknown"
        self.placemark = placemark
        self.serverModDate = serverModDate

        super.init()
    }
    
    public convenience init(id: String, name: String, coordinates: CLLocationCoordinate2D, postalAddress: CNPostalAddress? = nil, keywords: String? = nil, serverModDate: Date? = nil, flags: AddressFlags = []) {
        let placemark = (postalAddress != nil) ? MKPlacemark(coordinate: coordinates, postalAddress: postalAddress!) : MKPlacemark(coordinate: coordinates)
        
        self.init(id: id, name: name, placemark: placemark, keywords: keywords, serverModDate: serverModDate, flags: flags)
    }
    
    public override var hash: Int {
        var hasher = Hasher()
        
        hasher.combine(id)
        hasher.combine(flags.rawValue)
        hasher.combine(name.lowercased())
        
        return hasher.finalize()
    }

    public override func isEqual(_ object: Any?) -> Bool {
        guard let rhs = object as? AddressDTO else { return false }
        
        var retval = false
        
        // First check UUIDs
        retval = self.id == rhs.id
        
        if retval {
            // Check flags and name
            retval = (self.flags == rhs.flags) &&
                    (self.name.lowercased() == rhs.name.lowercased())
        }
        
        return retval
    }
}

extension AddressDTO: Comparable {
    // Compare by name, caseless
    public static func <(lhs: AddressDTO, rhs: AddressDTO) -> Bool {
        return lhs.name.lowercased() < rhs.name.lowercased()
    }
}

