//
//  AddressObject.swift
//  WherePrayerModel
//
//  Created by Jeff Shulman on 4/18/18.
//  Copyright © 2018 Lighthouse Woods LLC. All rights reserved.
//
//  Realm Address object

import Foundation
import MapKit
import RealmSwift

@objcMembers
public class AddressObject: Object, WPObjectProtocol {
    // MARK: Realm
    dynamic public var id = UUID().uuidString
    dynamic public var ckSystemFields: Data?
    dynamic public var addressString: String?
    dynamic public var flags: Int = 0
    dynamic public var keywords: String?
    dynamic public var mapImage: ImageObject?
    dynamic public var name: String = ""
    dynamic public var serverModDate: Date?

    // These get set via placemark
    @objc private dynamic var latitude: Double = 0.0
    @objc private dynamic var longitude: Double = 0.0

    @objc private dynamic var placemarkData: Data?

    public override var debugDescription: String {
        return """
            AddressObject
                id: \(id)
                ckSystemFields: \(String(describing: ckSystemFields))
                flags: \(String(describing: flags))
                addressString: \(String(describing: addressString))
                keywords: \(String(describing: keywords))
                mapImage: \(String(describing: mapImage))
                name: \(name))
                location: \(String(describing: location))
                placemark: \(String(describing: placemark))
                serverModDate: \(String(describing: serverModDate))\n
            """
    }

    public override static func primaryKey() -> String? {
        return "id"
    }
    
    public func myCopy() -> AnyObject {
        let addressObject = AddressObject(id: self.id, name: self.name)
        
        addressObject.ckSystemFields = self.ckSystemFields != nil ? Data(self.ckSystemFields!) : nil
        addressObject.flags = self.flags
        addressObject.addressString = self.addressString
        addressObject.keywords = self.keywords
        addressObject.mapImage = self.mapImage
        addressObject.latitude = self.latitude
        addressObject.longitude = self.longitude
        addressObject.placemarkData = self.placemarkData != nil ? Data(self.placemarkData!) : nil
        addressObject.serverModDate = self.serverModDate

        return addressObject
    }
    
    var coordinates: CLLocationCoordinate2D {
        get {
            return CLLocationCoordinate2D(latitude: latitude, longitude: longitude)
        }
        
        set {
            placemark = MKPlacemark(coordinate: newValue)
        }
    }
    
    var location: CLLocation {
        get {
            return CLLocation(latitude: latitude, longitude: longitude)
        }
        
        set {
            coordinates = CLLocationCoordinate2D(latitude: newValue.coordinate.latitude, longitude: newValue.coordinate.longitude)
        }
    }
    
    var placemark: CLPlacemark? {
        get {
            guard let placemarkData = placemarkData else {
                return nil
            }
            
            return try? NSKeyedUnarchiver.unarchivedObject(ofClass: CLPlacemark.self, from: placemarkData)
        }
        
        set {
            if let thePlacemark = newValue {
                placemarkData = try? NSKeyedArchiver.archivedData(withRootObject: thePlacemark, requiringSecureCoding: true)

                let coordinates = thePlacemark.location!.coordinate
                latitude = coordinates.latitude
                longitude = coordinates.longitude
            }
            else {
                placemarkData = nil
                latitude = 0
                longitude = 0
            }
        }
    }

    convenience init(id: String, name: String) {
        self.init()

        self.id = id
        self.name = name
    }

    convenience init(location: CLLocation, name: String) {
        self.init()
        
        self.location = location
        self.name = name
    }

    convenience init(coordinates: CLLocationCoordinate2D, name: String, serverModDate: Date? = nil) {
        self.init()
        
        self.coordinates = coordinates
        self.name = name
        self.serverModDate = serverModDate
    }

    convenience init(placemark: CLPlacemark, name: String, serverModDate: Date? = nil) {
        self.init()
        
        self.placemark = placemark
        self.name = name
        self.serverModDate = serverModDate
    }
    
    public override var hash: Int {
        var hasher = Hasher()
        
        hasher.combine(id)
        hasher.combine(flags)
        hasher.combine(addressString)
        hasher.combine(keywords)
        hasher.combine(coordinates)
        hasher.combine(placemarkData)

        return hasher.finalize()
    }
    
    public override func isEqual(_ object: Any?) -> Bool {
        guard let rhs = object as? AddressObject else { return false }
        
        if self.id != rhs.id {
            return false
        }
        
        if self.flags != rhs.flags {
            return false
        }
        
        if self.addressString != rhs.addressString {
            return false
        }
        
        if self.keywords != rhs.keywords {
            return false
        }
        
        if self.coordinates != rhs.coordinates {
            return false
        }
        
        if self.placemarkData != nil && rhs.placemarkData != nil {
            if self.placemarkData!.hashValue != rhs.placemarkData!.hashValue {
                return false
            }
        }
        
        return true
    }
    
    // MARK: - Exported
    
    // Called when this address is about to be deleted. Returns any ImageObject ID that were also deleted.
    // Must be called in a realm write block
    public func aboutToDeleteCascadeImages(in realm: RealmProtocol) -> String? {
        if mapImage == nil {
            return nil
        }
        
        let mapObject = mapImage!
        
        mapImage = nil
        
        if ImageObject.unreferencedPredicate().evaluate(with: mapObject) {
            let retval = mapObject.id
            
            try? mapObject.setImageData(nil)
            
            realm.delete(mapObject)
            
            return retval
        }
        
        return nil
    }
    
    // MARK: - Predicates
    
    /// Returns NSPredicate for Address by address id
    ///
    /// - parameter addressIDString: UUID string to look for
    public static func idPredicate( addressID: String ) -> NSPredicate {
        return NSPredicate(format: "id == %@", addressID as CVarArg)
    }
    
    /// Returns NSPredicate for Address's by address ids
    ///
    /// - parameter addressIDStrings: UUID strings to look for
    public static func idPredicate( addressIDs: [String] ) -> NSPredicate {
        return NSPredicate(format: "id IN %@", addressIDs as CVarArg)
    }
    
    /// Returns NSPredicate for Address by address name
    ///
    /// - parameter addressIDString: UUID string to look for
    public static func namePredicate( addressName: String ) -> NSPredicate {
        return NSPredicate(format: "name == %@", addressName as CVarArg)
    }
    
    /// Returns NSPredicate for Address that has the given flags
    ///
    /// - parameter flags: flags to look for
    public static func hasFlagsPredicate( flags: AddressFlags) -> NSPredicate {
        return NSPredicate(format: "flags == %i", flags.rawValue)
    }

    /// Returns NSPredicate for Address that does not have the given flags
    ///
    /// - parameter flags: flags to not look for
    public static func doesNotHaveFlagsPredicate( flags: AddressFlags) -> NSPredicate {
        return NSPredicate(format: "flags != %i", flags.rawValue)
    }
}
