//
//  ImageObject.swift
//  WherePrayerModel
//
//  Created by Jeff Shulman on 4/18/18.
//  Copyright © 2018 Lighthouse Woods LLC. All rights reserved.
//
//  Realm Image object

import Foundation
import RealmSwift

@objcMembers
public class ImageObject: Object, WPObjectProtocol {
    // MARK: Realm
    dynamic public var id = UUID().uuidString
    dynamic public var ckSystemFields: Data?
    dynamic var imageDataFileName: String?
    dynamic var serverModDate: Date?

    let addressesForMap = LinkingObjects(fromType: AddressObject.self, property: "mapImage")

    public override var debugDescription: String { return """
        ImageObject
            id: \(id)
            ckSystemFields: \(String(describing: ckSystemFields))
            imageDataFileName: \(String(describing: imageDataFileName))
            addressesForMap: \(addressesForMap))
            serverModDate: \(String(describing: serverModDate))\n
       """
    }

    public override static func primaryKey() -> String? {
        return "id"
    }
    
    public func myCopy() -> AnyObject {
        return ImageObject.init(value: self)
    }
    
    convenience init(id: String) {
        self.init()
        
        self.id = id
    }
    
    // MARK: - Exports
    
    class func fileURLForImage(id: String) throws -> URL {
        let directoryURL = FileManager.default.urls(for: .documentDirectory, in: .userDomainMask).first!.appendingPathComponent("images")
        try FileManager.default.createDirectory(at: directoryURL, withIntermediateDirectories: true, attributes: nil)
        let fileURL = directoryURL.appendingPathComponent(id)
        
        return fileURL
    }
    
    public func getImageData() throws -> Data? {
        if let fileName = imageDataFileName {
            let url = try ImageObject.fileURLForImage(id: fileName)
            return try Data(contentsOf: url)
        }
        else {
            return nil
        }
    }
    
    public func setImageData(_ imageData: Data?) throws {
        if imageData == nil {
            if imageDataFileName != nil {
                // Delete the existing file
                let url = try ImageObject.fileURLForImage(id: imageDataFileName!)
                try FileManager.default.removeItem(at: url)
            }
            
            imageDataFileName = nil
        }
        else {
            let url = try ImageObject.fileURLForImage(id: id)
            
            try imageData!.write(to: url)
            
            imageDataFileName = id
        }
    }
    
    // MARK: - Predicates
    
    /// Returns NSPredicate for Images's by image id
    ///
    /// - parameter imageIDString: UUID string to look for
    public static func idPredicate( imageID: String ) -> NSPredicate {
        return NSPredicate(format: "id == %@", imageID as CVarArg)
    }
    
    /// Returns NSPredicate for Images's by image ids
    ///
    /// - parameter imageIDStrings: UUID strings to look for
    public static func idPredicate( imageIDs: [String] ) -> NSPredicate {
        return NSPredicate(format: "id IN %@", imageIDs as CVarArg)
    }
    
    /// Returns NSPredicate for Images that are not references by any prayer/address
    public static func unreferencedPredicate() -> NSPredicate {
        return NSPredicate(format: "addressesForMap.@count == 0 AND privatePrayers.@count == 0 AND privatePrayersForMap.@count == 0 AND publicPrayersForMap.@count == 0")
    }
}

