//
//  LocalDataLayerRealmImpl.swift
//  WherePrayerModel
//
//  Created by Jeff Shulman on 4/18/18.
//  Copyright © 2018 Lighthouse Woods LLC. All rights reserved.
//
//  Realm implementation of the protocol for handling locally stored data

import Foundation
import CloudKit
import MapKit
import RealmSwift

final public class LocalDataLayerRealmImpl: LocalDataLayer {
    let timeoutValue : TimeInterval = 10.0
    
    // Set to indicate we are under unit testing
    public static var sUnitTesting = false
    
    // Changes
//    private var changeNotification: Signal<LocalDataLayerChangeNotification, Never>
//    private var changeNotificationObserver: Signal<LocalDataLayerChangeNotification, Never>.Observer

    private var addressObjects: Results<AddressObject>?
    private var addressToken: NotificationToken?
    private var imageObjects: Results<ImageObject>?
    private var imageToken: NotificationToken?

    private let inMemory: Bool
    private let logNotifications: Bool = false

    private let writeQueue = DispatchQueue(label: "com.lighthousewoods.wp.localDataQueue",
                                           qos: .utility,
                                           attributes: [])
    
    // To allow unit test code to wait for completion of operations
    private let writeSyncGroup = DispatchGroup()
    
    // For unit testing
    private var testRealm: RealmProtocol.Type?
    
    public init(inMemory: Bool, testRealm: RealmProtocol.Type? = nil) {
        self.inMemory = inMemory
        self.testRealm = testRealm
        
        let theRealm: RealmProtocol
        do {
            theRealm = try self.realm()
        }
        catch (let error) {
            fatalError("Could not initialize Realm \(error)")
        }

        installObservers(on: theRealm)
    }
    
    deinit {
        addressToken?.invalidate()
        addressToken = nil
    }
    
    // MARK: - Protocol methods
    
    public func asyncOperationsCompleted() -> Bool {
        let timeOut = writeSyncGroup.wait(wallTimeout: .now() + 0.5)
        return timeOut == .success
    }
    
    public func realm() throws -> RealmProtocol {
        let config = self.inMemory ?
                        Realm.Configuration(inMemoryIdentifier: "TemporaryRealm", deleteRealmIfMigrationNeeded: true) :
                        Realm.Configuration(deleteRealmIfMigrationNeeded: true)
        
        
        do {
            if let testRealm = testRealm {
                let realm = try testRealm.init(configuration: config)
                
                return realm
            }
            else {
                let realm = try Realm(configuration: config)
                
                return realm
            }
        }
        catch let error {
            throw LocalDataLayerError.noDatabase(error)
        }
    }
    
    public func countItems( matching predicate:NSPredicate?, ofType:AnyClass ) -> Result<Int, Swift.Error> {
        guard let objectType = ofType as? Object.Type else {
            fatalError("Trying to count a non Realm object \(ofType)")
        }
        
        do {
            let realm = try self.realm()
            
            var items = realm.objects(objectType)
            
            if predicate != nil {
                items = items.filter(predicate!)
            }
            
            return .success(items.count)
        }
        catch let error {
            return .failure(LocalDataLayerError.countItemsFailed(error))
        }
    }
    
    @discardableResult
    public func clearDataStore() -> Result<Void, Swift.Error> {
        do {
            let realm = try self.realm()
            
            try realm.write(withoutNotifying: []) {
                realm.deleteAll()
            }
            
            // Special delete
//            postDeleteNotifier(changeType: .privatePrayer, deletedObjectIds: ["-1"])

            return .success(())
        }
        catch let error {
            return .failure(LocalDataLayerError.clearFailed(error))
        }
    }
    
    // MARK: - Utilities
    
    private func installObservers(on realm: RealmProtocol) {
        addressObjects = realm.objects(AddressObject.self)
        addressToken = addressObjects!.observe { (changes) in
//            self?.changeNotifier(changeType: .address, changes: changes)
        }
        
        imageObjects = realm.objects(ImageObject.self)
        imageToken = imageObjects!.observe { (changes) in
//            self?.changeNotifier(changeType: .image, changes: changes)
        }
    }
    
}

